use axum::extract::ws::WebSocket;
use axum::http::StatusCode;
use reqwest::Url;
use serde::{Deserialize, Serialize};
use thiserror::Error;
use uuid::Uuid;

use crate::REQWEST_CLIENT;
use packet::{
    ClientboundPacket, HandshakeRequest, HandshakeResponse, PacketIo, PacketIoError,
    ServerboundPacket, WritePacket,
};

pub mod packet;

#[derive(Serialize, Deserialize)]
pub struct GameProfileProperty {
    pub name: String,
    pub value: String,
    pub signature: Option<String>,
}

#[derive(Serialize, Deserialize)]
pub struct GameProfile {
    pub id: Uuid,
    pub name: String,
    pub properties: Vec<GameProfileProperty>,
}

pub async fn authenticate_socket(
    socket: &mut WebSocket,
) -> Result<GameProfile, AuthenticateSocketError> {
    let join_id = Uuid::new_v4();

    socket
        .write_packet(&ClientboundPacket::HandshakeRequest(HandshakeRequest {
            join_id,
        }))
        .await?;

    let HandshakeResponse { player_name } = match socket
        .read_packet()
        .await?
        .ok_or(AuthenticateSocketError::ClientDisconnected)?
    {
        ServerboundPacket::HandshakeResponse(response) => response,
        #[allow(unreachable_patterns)]
        packet => return Err(AuthenticateSocketError::InvalidPacket(packet)),
    };

    let url = Url::parse_with_params(
        "https://sessionserver.mojang.com/session/minecraft/hasJoined",
        &[
            ("username", player_name.clone()),
            ("serverId", format!("ssi-{}", join_id)),
        ],
    )?;

    let response = REQWEST_CLIENT.get(url).send().await?;

    if !response.status().is_success() {
        return Err(AuthenticateSocketError::Non200Response(response.status()));
    }

    response.json().await.map_err(Into::into)
}

#[derive(Debug, Error)]
pub enum AuthenticateSocketError {
    #[error("non-200 status code: {0}")]
    Non200Response(StatusCode),
    #[error("invalid packet: {0:?}")]
    InvalidPacket(ServerboundPacket),
    #[error("client disconnected")]
    ClientDisconnected,
    #[error(transparent)]
    PacketIo(#[from] PacketIoError),
    #[error(transparent)]
    Reqwest(#[from] reqwest::Error),
    #[error(transparent)]
    UrlParse(#[from] url::ParseError),
}
