use axum::extract::ws::{Message, WebSocket};
use futures_util::{
    stream::{SplitSink, SplitStream},
    SinkExt, TryStreamExt,
};
use thiserror::Error;

use super::{ClientboundPacket, ServerboundPacket};

#[derive(Debug, Error)]
pub enum PacketIoError {
    #[error("the client has disconnected")]
    Disconnectetd,
    #[error("the socket is closing")]
    SocketClosing,
    #[error("the client sent an invalid websocket message: {0:?}")]
    InvalidMessage(Message),
    #[error(transparent)]
    Axum(#[from] axum::Error),
    #[error(transparent)]
    SerdeJson(#[from] serde_json::Error),
}

pub trait PacketIo {
    async fn read_packet(&mut self) -> Result<Option<ServerboundPacket>, PacketIoError>;
}

impl PacketIo for WebSocket {
    async fn read_packet(&mut self) -> Result<Option<ServerboundPacket>, PacketIoError> {
        process_packet(self.try_next().await?.ok_or(PacketIoError::Disconnectetd)?)
    }
}

impl PacketIo for SplitStream<WebSocket> {
    async fn read_packet(&mut self) -> Result<Option<ServerboundPacket>, PacketIoError> {
        process_packet(self.try_next().await?.ok_or(PacketIoError::Disconnectetd)?)
    }
}

fn process_packet(message: Message) -> Result<Option<ServerboundPacket>, PacketIoError> {
    match message {
        Message::Text(text) => Ok(Some(serde_json::from_str(&text)?)),
        Message::Ping(_) | Message::Pong(_) => Ok(None),
        Message::Close(_) => Err(PacketIoError::SocketClosing),
        msg => Err(PacketIoError::InvalidMessage(msg)),
    }
}

pub trait WritePacket {
    async fn write_packet(&mut self, packet: &ClientboundPacket) -> Result<(), PacketIoError>;
}

impl WritePacket for WebSocket {
    async fn write_packet(&mut self, packet: &ClientboundPacket) -> Result<(), PacketIoError> {
        let serialized = serde_json::to_string(packet)?;
        self.send(Message::Text(serialized)).await?;
        Ok(())
    }
}

impl WritePacket for SplitSink<WebSocket, Message> {
    async fn write_packet(&mut self, packet: &ClientboundPacket) -> Result<(), PacketIoError> {
        let serialized = serde_json::to_string(packet)?;
        self.send(Message::Text(serialized)).await?;
        Ok(())
    }
}
